import os
# import io
import sys
import requests

from tqdm import tqdm

import logging
logger = logging.getLogger(__name__)


# ########### URL ##############
def download_from_url(url, dst):
    """
    Download file with progressbar

    @param: url to download file
    @param: dst place to put the file
    """
    chunk_size = 1024  # 1048576

    try:
        file_size = int(requests.head(url).headers["Content-Length"])
    except:
        logger.warning("Could not access URL " + url)
        sys.exit(1)

    first_byte = os.path.getsize(dst) if os.path.exists(dst) else 0

    if first_byte >= file_size:
        logger.warning("Looks like the file has been downloaded already: " + dst)
        return file_size

    header = {"Range": "bytes=%s-%s" % (first_byte, file_size)}
    try:
        r = requests.get(url, headers=header, stream=True)
        file_size = int(r.headers['Content-Length'])
    except:
        logger.warning("Could not access URL " + url)
        sys.exit(1)

    def generate(raw, chunk_size):
        while True:
            chunk = raw.read(chunk_size)
            if not chunk:
                break
            yield chunk

    # need to access raw stream because r.iter_content() deflates .gz automatically
    # opening file in append mode
    with open(dst, 'ab') as fp:
        for chunk in tqdm(
                generate(r.raw, chunk_size=chunk_size),
                initial=first_byte // chunk_size,
                total=file_size // chunk_size,
                unit='KB',
                desc=dst,
                leave=True):
            fp.write(chunk)
    return file_size


def fetch_genome(name):
    # ftp://hgdownload.cse.ucsc.edu/goldenPath/hg19/bigZips/hg19.2bit
    # http://hgdownload.cse.ucsc.edu/goldenPath/hg19/bigZips/hg19.2bit
    assembly = name.rsplit('.', 1)[0] if '.2bit' in name else name
    assembly = assembly.lower()
    mode = 'http'

    url = f'{mode}://hgdownload.cse.ucsc.edu/goldenPath/{assembly}/bigZips/{assembly}.2bit'
    dst = f'{assembly}.2bit'
    download_from_url(url, dst)


def fetch_cohorts():
    url = "https://www.ncbi.nlm.nih.gov/research/mutagene/static/data/cohorts.tar.gz"
    dst = "cohorts.tar.gz"
    download_from_url(url, dst)


def fetch_examples():
    url = "https://www.ncbi.nlm.nih.gov/research/mutagene/static/data/"
    files = ["sample1.maf", "sample2.vcf"]
    for fname in files:
        download_from_url(url + fname, fname)


def fetch_MSKCC(dataset):
    """ Download dataset from cBioPortal https://www.cbioportal.org/datasets """
    url = "http://download.cbioportal.org/{}.tar.gz".format(dataset)
    dst = "{}.tar.gz".format(dataset)
    download_from_url(url, dst)



# from collections import defaultdict
# from mutagene.dna import complementary_nucleotide

import numpy as np
from numpy.random import multinomial
from sklearn.utils import resample

from mutagene.io.profile import write_profile_file, get_profile_attributes_dict
from mutagene.io.mutations_profile import read_auto_profile

import logging
logger = logging.getLogger(__name__)


def calc_profile(infile, outfile, genome, fmt='auto'):
    all_mutations = {}
    for f in infile:
        mutations, processing_stats = read_auto_profile(f, fmt=fmt, asm=genome)
        msg = "Loaded {} mutations".format(processing_stats['loaded'])
        if processing_stats['skipped'] > 0:
            msg += " skipped {} mutations due to mismatches with the reference genome".format(processing_stats['skipped'])
        logger.info(msg)
        all_mutations = {k: all_mutations.get(k, 0) + mutations.get(k, 0) for k in set(all_mutations) | set(mutations)}
    if sum(all_mutations.values()) == 0:
        logger.warn('Can not create profile')
        return
    profile = get_mutational_profile(all_mutations, counts=True)
    write_profile_file(outfile, profile)


def get_mutational_profile(mutational_profile_dict, counts=False):
    attrib = get_profile_attributes_dict()
    values = []
    total_mut_number = sum(mutational_profile_dict.values())
    for i, attr in enumerate(attrib):
        number = mutational_profile_dict.get(attr['context'] + attr['mutation'], 0)
        # freq = 0.000001 * number / total_mut_number
        if counts:
            freq = number
        else:
            freq = number / float(total_mut_number)
        # trinucleotide = attr['context'][0] + attr['mutation'][0] + attr['context'][1]
        # trinucleotide_freq = exome_trinucleotide_freq[trinucleotide]
        # values.append(3.0 * freq / trinucleotide_freq)
        values.append(freq)
    return values


def get_multisample_mutational_profile(samples_mutations, counts=False):
    samples_profiles = {}

    for sample, mutations in samples_mutations.items():
        samples_profiles[sample] = get_mutational_profile(mutations, counts)

    return samples_profiles


def generate_resampled_profiles(profile, k):
    profile = np.array(profile)
    N = np.sum(profile)
    new_profiles = multinomial(N, profile / N, size=k)
    for i in range(k):
        yield new_profiles[i]





    # @manager.command
    # def dump_nmf_decomp():
    #     for AB in ('A', 'B'):
    #         deconv = db.query(CachedObjects).get('deconvolution_' + AB).value

    #         # load samples in 5D or 10D
    #         W = deconv['W']
    #         H = deconv['H'].T
    #         ids = deconv['ids']

    #         for i, ID in enumerate(ids):
    #             profile = db.query(Signature).get(int(ID))
    #             v = np.array([get_signature_values_list(profile)]).ravel()
    #             x = np.array(H[i]).squeeze()
    #             residuals = np.linalg.norm(W.dot(x) - v)  # Frobenius norm
    #             N = '5' if AB == 'A' else '10'
    #             oname = "/Users/gonceare/projects/RA/ALL.decompose/{}_NMF{}.txt".format(ID, N)
    #             with open(oname, 'w') as o:
    #                 for j in range(H.shape[1]):
    #                     name = "MUTAGENE {}.{}".format(AB, j + 1)
    #                     o.write("{}\t{}\n".format(name, H[i, j]))
    #                 # print(ID, residuals)
    #                 o.write("Residuals\t{}\n".format(residuals))





import json
import os
import pandas as pd

import logging
logger = logging.getLogger(__name__)


def write_motif_matches(outfile, motif_matches):
    """Convert motif matches list of dictionaries to CSV format"""
    if len(motif_matches) > 0:
        header = motif_matches[0].keys()
        d = pd.DataFrame.from_records(motif_matches, columns=header).sort_values(by='pvalue')
        d.to_csv(outfile, sep="\t", index=False, float_format='%g')


def get_known_motifs():
    """Load a pre-bundled set of motifs from data directory"""
    dirname = os.path.dirname(os.path.realpath(__file__))
    fname = dirname + "/../data/motifs/motifs.json"
    with open(fname, 'r') as f:
        motifs = json.load(f)
    return motifs



# import csv
import numpy as np
from scipy import stats
from collections import defaultdict
import pandas as pd


def _get_stats(results):
    """
        Convert dictionary decomposition results into a pandas dataframe compatible with bootstrap stats results
    """
    if type(results) is np.ndarray:
        h = results
    else:
        exposure_dict = {x['name']: x['score'] for x in results if x['mutations'] != ''}
        mutations_dict = {x['name']: x['mutations'] for x in results if x['mutations'] != ''}
        signatures = list(exposure_dict.keys())

        exposure = [exposure_dict[name] for name in signatures]
        mutations = [mutations_dict[name] for name in signatures]

        h = np.array(exposure)
        m = np.array(mutations, np.int)

    df = pd.DataFrame(
        dict(
            zip(
                "signature\texposure\tmutations".split(),
                [signatures, h, m])))
    return df.sort_values(by=["mutations", "exposure"], ascending=False)


def _get_bootstrap_stats_percentile(bootstrap_results, level):
    """
        level is converted to percentile: level 90 means an interval 9% - 95%

        Quantile-based bootstrap intervals
        Returns pandas data frame - each row is a signature, zeros not removed
    """
    ci_high = (100.0 + level) / 2.0
    ci_low = (100.0 - level) / 2.0

    exposures_lists = defaultdict(list)
    mutations_lists = defaultdict(list)
    signatures = set()

    for results in bootstrap_results:
        for x in results:
            if x['mutations'] != '' and x['score'] != '':
                exposures_lists[x['name']].append(x['score'])
                mutations_lists[x['name']].append(x['mutations'])
                signatures |= set([x['name'], ])
    signatures = list(signatures)

    # h: exposures, float [0..1]
    exposures = np.array([exposures_lists[name] for name in signatures]).T

    h = np.percentile(exposures, 50.0, axis=0)  # median
    h_ci_low = np.percentile(exposures, ci_low, axis=0)
    h_ci_high = np.percentile(exposures, ci_high, axis=0)

    # m: mutations, integer [0, 1, 2...]
    mutations = np.array([mutations_lists[name] for name in signatures]).T
    m = np.round(np.percentile(mutations, 50.0, axis=0))
    m_ci_low = np.round(np.percentile(mutations, ci_low, axis=0))
    m_ci_high = np.round(np.percentile(mutations, ci_high, axis=0))

    df = pd.DataFrame(
        dict(
            zip(
                "signature\texposure\tmutations\texposure_low\texposure_high\tmutations_low\tmutations_high".split(),
                [signatures, h, m, h_ci_low, h_ci_high, m_ci_low, m_ci_high])))
    return df.sort_values(by=["mutations", "exposure"], ascending=False)


def _get_bootstrap_stats_t(sample_results, bootstrap_results, n, level):
    """
        Calculate approximate bias-corrected t-based confidence intervals for a list of samples using Manly (2007) method.

        n is the number of mutations

        t-based confidence intervals for given confidence level will be based on t distribution with n-1 degrees of freedom

        bias-corrected means are calculated as 2 * theta - mean(bootstrap_theta), where:
            - theta is exposure calculated without the bootstrap
            - mean(bootstrap_theta) is expected value of exposure in bootstrap samples

        Returns pandas data frame - each row is a signature, zeros not removed
    """
    ci_level = (1.0 + level / 100.0) / 2.0

    # ppf: Percent point function (inverse of cdf — percentiles) of t-distribution with (n-1) d.f.
    t = stats.t.ppf(ci_level, n - 1)

    exposure = defaultdict(float)
    mutations = defaultdict(float)
    exposures_lists = defaultdict(list)
    mutations_lists = defaultdict(list)
    signatures = set()

    for x in sample_results:
        if x['mutations'] != '' and x['score'] != '':
            exposure[x['name']] = x['score']
            mutations[x['name']] = x['mutations']

    for results in bootstrap_results:
        for x in results:
            if x['mutations'] != '' and x['score'] != '':
                exposures_lists[x['name']].append(x['score'])
                mutations_lists[x['name']].append(x['mutations'])
                signatures |= set([x['name'], ])
    signatures = list(signatures)

    # h: exposures, float [0..1]
    h = np.array([2 * exposure[name] - np.nanmean(exposures_lists[name]) for name in signatures])
    h_sem = np.array([stats.sem(exposures_lists[name]) for name in signatures])
    h_ci_low = np.clip(h - t * h_sem, a_min=0.0, a_max=None)
    h_ci_high = np.clip(h + t * h_sem, a_min=None, a_max=1.0)
    h = np.clip(h, a_min=0.0, a_max=None)

    # m: mutations, integer [0, 1, 2...]
    m = np.array([2 * mutations[name] - np.nanmean(mutations_lists[name]) for name in signatures])
    m_sem = np.array([stats.sem(mutations_lists[name]) for name in signatures])
    m_ci_low = np.clip(m - t * m_sem, a_min=0, a_max=None)
    m_ci_high = m + t * m_sem
    m = np.clip(m, a_min=0, a_max=None)

    df = pd.DataFrame(
        dict(
            zip(
                "signature\texposure\tmutations\texposure_low\texposure_high\tmutations_low\tmutations_high".split(),
                [signatures, h, m, h_ci_low, h_ci_high, m_ci_low, m_ci_high]
            )))
    return df.sort_values(by="exposure", ascending=False)


def write_decomposition(
    fname, samples_results, signature_ids, mutations_threshold=0,
    bootstrap_method=None, profile=None,
    bootstrap_results=None, bootstrap_level=None
):
    """
        Process sample results for the input data and (optionally) bootstrapped data
        and save a resulting table with or without the optional bootstrap columns
    """
    bootstrap = bootstrap_results is not None and bootstrap_level is not None and profile is not None

    # print("Bootstrap", bootstrap)
    dfs = []
    for sample in samples_results.keys():
        if bootstrap:
            if bootstrap_method == 't':
                # t-distribution-based confidence intervals
                p = profile[sample]
                n = int(np.sum(p))
                assert n > 1
                df_sample = _get_bootstrap_stats_t(samples_results[sample], bootstrap_results[sample], n, bootstrap_level)
            elif bootstrap_method == 'p':
                # percentile-based confidence intervals
                df_sample = _get_bootstrap_stats_percentile(bootstrap_results[sample], bootstrap_level)
            else:
                raise ValueError("Incorrect bootstrap_method value, only 't' or 'p' are recognized")
        else:
            df_sample = _get_stats(samples_results[sample])
        df_sample.insert(0, 'sample', sample)
        dfs.append(df_sample)

    # join results for multiple samples
    if len(dfs) == 0:
        return
    df = pd.concat(dfs)

    # convert mutations to int
    df = df.astype({'mutations': 'int32'})
    if 'mutations_low' in df:
        df = df.astype({'mutations_high': 'int32', 'mutations_low': 'int32'})

    # filter mutations by threshold
    df = df[df.mutations > mutations_threshold]

    if df is not None:
        df.to_csv(fname, sep="\t", index=False, float_format='%g')


def read_decomposition(fname):
    """
        Load results of signature decomposition from file (filename)
        Return tuple: numpy array with exposures and a list of signare names
    """
    signature_ids = []
    h = []

    try:
        with open(fname) as f:
            for line in f:
                a, b = line.strip().split()
                signature_ids.append(a)
                h.append(float(b))
    except:
        return None, None
    # except FileNotFoundError:
    #     return None, None

    return np.array(h), signature_ids




import json
from subprocess import Popen, PIPE

try:
    from .identify_signature import *
    # from . import *
    from .mutations import *
    from .json_structs import *
except:
    pass

from itertools import chain, islice

from collections import namedtuple, defaultdict
import gzip
import csv
import os
# from .COSMIC import iter_CLP_mutations


TWOBIT_GENOMES_PATH = '/net/pan1/mutagene/data/genomes/'


def make_NCI60_report(prefix, n=30, decomposition='L'):
    s = {5: 'MUTAGENE A.', 10: 'MUTAGENE B.', 30: 'COSMIC '}[n]

    COSMIC_CLINICAL = prefix + "/cellline_nci60_clinical_data.tsv"

    stats_columns = ("N", "Processed_C", "Processed_NC", "N_C", "N_NC", "Skipped_C", "Skipped_NC", "Skipped_indels_C", "Skipped_indels_NC", "Skipped_nucleotide_C", "Skipped_nucleotide_NC", "Skipped_context_C", "Skipped_context_NC")
    with open(COSMIC_CLINICAL) as f, open(prefix + "/report_{}{}.txt".format(decomposition, n), 'w') as o:
        o.write("SAMPLE_ID\tCANCER_TYPE\t" + "\t".join(stats_columns) + "\t" + "\t".join(["{}{}".format(s, x) for x in range(1, n + 1)]) + "\tOther\tUnexplained\n")
        for line in islice(f, 1, None):
            if line.startswith('#'):
                continue
            if line.startswith('SAMPLE_ID'):
                continue
            fields = line.split("\t")
            if len(fields) < 3:
                continue
            SAMPLE_ID = fields[0]
            CANCER = fields[3]

            if not os.path.isfile(prefix + "/profiles/{}.stats".format(SAMPLE_ID)):
                continue

            print(SAMPLE_ID, CANCER)

            o.write("{}\t{}\t".format(SAMPLE_ID, CANCER))

            stats_values = {}
            with open(prefix + "/profiles/{}.stats".format(SAMPLE_ID)) as stats:
                for line in stats:
                    stats, value = line.split("\t")
                    stats_values[stats] = value.strip()

            for stats in stats_columns:
                o.write("{}\t".format(stats_values[stats]))

            exposure = defaultdict(float)
            with open(prefix + "/decompose/{}_{}-{}.txt".format(SAMPLE_ID, decomposition, n)) as d:
                for line in d:
                    signature, value = line.split("\t")
                    value = float(value)
                    if signature.startswith(s):
                        exposure[signature] = value
                    if signature.startswith("Other"):
                        exposure['other'] = value
            unexplained = 1.0 - sum(exposure.values())
            o.write("\t".join(["{0:.2f}".format(exposure["{}{}".format(s, x)]) for x in range(1, n + 1)]) + "\t{:.2f}\t{:.2f}\n".format(exposure['other'], unexplained))


def read_MAF_extended(muts, asm=None):
    cn = complementary_nucleotide
    N_skipped = 0

    samples = defaultdict(list)

    raw_mutations = []
    for i, line in enumerate(muts.split("\n")):
        if line.startswith("#"):
            continue
        if line.startswith("Hugo_Symbol"):
            continue
        if len(line) < 10:
            continue

        col_list = line.split("\t")
        # for a, b in enumerate(col_list):
        #     print(a, b)

        if len(col_list) < 13:
            continue
        try:
            # assembly_build = col_list[3]  # MAF ASSEMBLY
            # strand = col_list[7]    # MAF STRAND
            # ID = col_list[2]

            # chromosome is expected to be one or two number or one letter
            chrom = col_list[4]  # MAF CHROM
            if chrom.lower().startswith("chr"):
                chrom = chrom[3:]
            # if len(chrom) == 2 and chrom[1] not in "0123456789":
            #     chrom = chrom[0]

            pos = int(col_list[5])      # MAF POS START
            pos_end = int(col_list[6])  # MAF POS END
            x = col_list[10]            # MAF REF
            y1 = col_list[11]           # MAF ALT1
            y2 = col_list[12]           # MAF ALT2
        except:
            # raise
            continue

        sample = None
        try:
            sample = col_list[15]           # Tumor_Sample_Barcode
        except:
            raise

        if pos != pos_end:
            continue

        # skip if found unexpected nucleotide characters
        if len(set([x, y1, y2]) - set(nucleotides)) > 0:
            continue

        y = y1 if y1 != x else None
        y = y2 if y2 != x else y
        if y is None:
            continue

        samples[sample].append((chrom, pos, x, y))

    processing_stats = defaultdict(dict)

    MAX = 100000000
    for sample, raw_mutations in samples.items():
        print("sample", sample, len(raw_mutations))

        # if sample != "786_0":
        #     break

        mutations = defaultdict(float)
        if len(raw_mutations) > 0:
            if len(raw_mutations) > MAX:
                raw_mutations = raw_mutations[:MAX]

            contexts = get_context_batch(raw_mutations, asm)

            if contexts is None:
                return None, None

            if len(contexts) == 0:
                return None, None

            for (chrom, pos, x, y) in raw_mutations:
                p5, p3 = contexts.get((chrom, pos), ("N", "N"))

                if len(set([p5, x, y, p3]) - set(nucleotides)) > 0:
                    # print("Skipping invalid nucleotides")
                    N_skipped += 1
                    continue

                if x in "CT":
                    mutations[p5 + p3 + x + y] += 1.0
                else:
                    # complementary mutation
                    mutations[cn[p3] + cn[p5] + cn[x] + cn[y]] += 1.0

        N_loaded = int(sum(mutations.values()))
        processing_stats[sample] = {'loaded': N_loaded, 'skipped': N_skipped, 'format': 'MAF'}
        samples[sample] = mutations

    return samples, processing_stats


uniq_mut = defaultdict(set)


def read_CLP_extended(f, asm=None, ids=None):
    cn = complementary_nucleotide
    # N_skipped = 0
    # uniq_mut = defaultdict(set)

    samples = defaultdict(list)
    samples_raw = defaultdict(list)
    samples_mutations = defaultdict(int)
    samples_skipped_indels = defaultdict(int)
    samples_skipped_nucleotide = defaultdict(int)
    samples_skipped_context = defaultdict(int)

    CLPRecord = namedtuple('CLPRecord',
        'gene, accession, CDS_length, HGNC, sample_name, sample_id, tumor_id, site, site_subtype1, site_subtype2, site_subtype3, ' +
        'histology, histology_subtype1, histology_subtype2, histology_subtype3, GWS, mut_id, mut_CDS, mut_AA, mut_descr, mut_zygosity, LOH, ' +
        'GRCh, GRC_pos, GRC_strand, SNP, FATHMM, FATHMM_score, somatic, verification, PMID, study_id, institute, institute_address, catalogue_no, ' +
        'sample_source, tumor_origin, age')

    raw_mutations = []
    for rec in map(CLPRecord._make, csv.reader(islice(f, 1, None), delimiter='\t')):
        # print rec.gene
        # print rec.mut_CDS
        # print rec.mut_AA
        # print rec.somatic

        try:
            sample = rec.sample_name.replace("-", "_").upper()

            chrom = ''
            pos = 0
            pos_end = 0
            if len(rec.GRC_pos) > 0:
                chrom, start_end = rec.GRC_pos.split(":")
                pos, pos_end = start_end.split("-")
                pos = int(pos)
                pos_end = int(pos_end)

            x = rec.mut_CDS[-3]
            y = rec.mut_CDS[-1]

            if ids and sample not in ids:
                continue

            mut = (rec.GRC_pos if len(rec.GRC_pos) > 0 else rec.mut_id, x, y)
            if mut in uniq_mut[sample]:
                # It's probably another isoform!
                # print("SKIPPING DUPLICATE", sample, rec.GRC_pos, x, y)
                continue
            uniq_mut[sample] |= set([mut])

            samples_mutations[sample] += 1

        except Exception as e:
            # raise
            print("SKIPPING RECORD", str(e), rec)
            continue

        if pos != pos_end or "ins" in rec.mut_CDS or "del" in rec.mut_CDS:
            samples_skipped_indels[sample] += 1
            continue

        if len(chrom) == 0:
            print("SKIPPING. NO GRCh specified", rec)
            samples_skipped_nucleotide[sample] += 1
            continue

        # skip if found unexpected nucleotide characters
        if len(set([x, y]) - set(nucleotides)) > 0:
            samples_skipped_nucleotide[sample] += 1
            continue

        samples[sample].append((chrom, pos, x, y))

    processing_stats = defaultdict(dict)

    # MAX = 100000000
    for sample, raw_mutations in samples.items():
        print("sample", sample, len(raw_mutations))

        # if sample != "786_0":
        #     break

        mutations = defaultdict(float)
        if len(raw_mutations) > 0:
            # if len(raw_mutations) > MAX:
            #     raw_mutations = raw_mutations[:MAX]

            contexts = get_context_batch(raw_mutations, asm)

            if contexts is None:
                return None, None

            if len(contexts) == 0:
                return None, None

            for (chrom, pos, x, y) in raw_mutations:
                p5, p3 = contexts.get((chrom, pos), ("N", "N"))

                if len(set([p5, x, y, p3]) - set(nucleotides)) > 0:
                    print("Skipping invalid nucleotides {}:{} {}[{}>{}]{}".format(chrom, pos, p5, x, y, p3))
                    # N_skipped += 1
                    samples_skipped_context[sample] += 1
                    continue

                if x in "CT":
                    mutations[p5 + p3 + x + y] += 1.0
                else:
                    # complementary mutation
                    mutations[cn[p3] + cn[p5] + cn[x] + cn[y]] += 1.0

        N = samples_mutations[sample]
        N_loaded = int(sum(mutations.values()))
        processing_stats[sample] = {'processed': N, 'loaded': N_loaded, 'skipped': N - N_loaded, 'skipped_indels': samples_skipped_indels[sample], 'skipped_context': samples_skipped_context[sample], 'skipped_nucleotide': samples_skipped_nucleotide[sample] }
        samples[sample] = mutations
        samples_raw[sample] = raw_mutations

    return samples, processing_stats, samples_raw


def read_CLP_NCV(f, asm=None, ids=None):
    cn = complementary_nucleotide
    # N_skipped = 0

    # uniq_mut = defaultdict(set)
    samples = defaultdict(list)
    samples_raw = defaultdict(list)
    samples_mutations = defaultdict(int)
    samples_skipped_nucleotide = defaultdict(int)
    samples_skipped_context = defaultdict(int)
    samples_skipped_indels = defaultdict(int)

    CLPNCVRecord = namedtuple('CLPNCVRecord',
        'sample_name, sample_id, tumor_id, site, site_subtype1, site_subtype2, site_subtype3, ' +
        'histology, histology_subtype1, histology_subtype2, histology_subtype3, GWS, mut_id, mut_zygosity, ' +
        'GRCh, GRC_pos, somatic, WT, MUT, SNP, FATHMM_score, FATHMM_groups, FATHMM_MKL_score, FATHMM_MKL_groups, genome, exome, study')

    raw_mutations = []
    for rec in map(CLPNCVRecord._make, csv.reader(islice(f, 1, None), delimiter='\t')):
        try:
            sample = rec.sample_name.replace("-", "_").upper()

            chrom, start_end = rec.GRC_pos.split(":")
            pos, pos_end = start_end.split("-")
            pos = int(pos)
            pos_end = int(pos_end)

            x = rec.WT
            y = rec.MUT

            if ids and sample not in ids:
                continue

            samples_mutations[sample] += 1

            mut = (rec.GRC_pos, x, y)
            if mut in uniq_mut[sample]:
                # It's probably another isoform!
                print("SKIPPING DUPLICATE", sample, rec.GRC_pos, x, y)
                continue
            uniq_mut[sample] |= set([mut])

        except Exception as e:
            # raise
            print("SKIPPING RECORD", str(e), rec)
            continue

        if pos != pos_end or len(x) != 1 or len(y) != 1:
            # print("SKIPPING INDEL", pos, pos_end, x, y)
            samples_skipped_indels[sample] += 1
            continue

        # skip if found unexpected nucleotide characters
        if len(set([x, y]) - set(nucleotides)) > 0:
            # print("SKIPPING UNEXPECTED", x, y)
            samples_skipped_nucleotide[sample] += 1
            continue

        samples[sample].append((chrom, pos, x, y))

    processing_stats = defaultdict(dict)

    # MAX = 100000000
    for sample, raw_mutations in samples.items():
        print("sample", sample, len(raw_mutations))

        # if sample != "786_0":
        #     break

        mutations = defaultdict(float)
        if len(raw_mutations) > 0:
            # if len(raw_mutations) > MAX:
            #     raw_mutations = raw_mutations[:MAX]

            contexts = get_context_batch(raw_mutations, asm)

            if contexts is None:
                return None, None

            if len(contexts) == 0:
                return None, None

            for (chrom, pos, x, y) in raw_mutations:
                p5, p3 = contexts.get((chrom, pos), ("N", "N"))

                if len(set([p5, x, y, p3]) - set(nucleotides)) > 0:
                    # print("Skipping invalid nucleotides")
                    print("Skipping invalid nucleotides {}[{}>{}]{}".format(p5, x, y, p3))
                    # N_skipped += 1
                    samples_skipped_context[sample] += 1
                    continue

                if x in "CT":
                    mutations[p5 + p3 + x + y] += 1.0
                else:
                    # complementary mutation
                    mutations[cn[p3] + cn[p5] + cn[x] + cn[y]] += 1.0

        N = samples_mutations[sample]
        N_loaded = int(sum(mutations.values()))
        # processing_stats[sample] = {'loaded': N_loaded, 'skipped': N - N_loaded}
        processing_stats[sample] = {'processed': N, 'loaded': N_loaded, 'skipped': N - N_loaded, 'skipped_indels': samples_skipped_indels[sample], 'skipped_context': samples_skipped_context[sample], 'skipped_nucleotide': samples_skipped_nucleotide[sample] }
        samples[sample] = mutations
        samples_raw[sample] = mutations_raw

    return samples, processing_stats, samples_raw



def deconstruct_sigs(profile_fname, sample):
    script = """
library(jsonlite)
library(deconstructSigs)
s <- t(read.table('{}', sep="\t", header=FALSE, row.names=1))
row.names(s) <- '{}'
s <- as.data.frame(s)
w <- whichSignatures(s/sum(s), signatures.ref=signatures.cosmic)
toJSON(w)
"""
    script = script.format(profile_fname, sample).encode("utf-8")
    proc = Popen(["/Users/gonceare/anaconda3/envs/mutagene/bin/Rscript", "-"], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    out, err = proc.communicate(script)
    # exitcode = proc.returncode
    # print(exitcode, out, err)
    json_string = out.decode("utf-8")
    w = json.loads(json_string)
    # pprint.pprint(w)
    result = []
    for k, v in w['weights'][0].items():
        if k.startswith('_row'):
            continue
        if float(v) == 0.0:
            continue
        result.append({
            'name': k.replace('Signature.', 'COSMIC '),
            'score': v})
    return None, result

# GRC37
# from cosmic website 468 noncoding + 330 coding = 798
# grep files: 519 + 350
# 818 loaded 51 skipped = 869

# GRC38
# ??? 394
# 321 + 461 = 782
# grep 344 + 511 = 855
#OLD ISO: BT_549': {'loaded': 805, 'skipped': 50} = 855
# 'BT_549': {'loaded': 714, 'skipped': 36}: = 750
#


def export_maf(prefix, maf_file):
    """Create a MAF file with mutations

    Mutations in NCI-60 cell lines are extracted from COSMIC CLP

    Arguments:
        prefix {string} -- Location of COSMIC data
        maf_file {string} -- output file name

    Returns:
        None
    """

    """
    GDC MAF Format v.1.0.0
    From https://docs.gdc.cancer.gov/Data/File_Formats/MAF_Format/

    1 - Hugo_Symbol   HUGO symbol for the gene (HUGO symbols are always in all caps). "Unknown" is used for regions that do not correspond to a gene
    2 - Entrez_Gene_Id    Entrez gene ID (an integer). "0" is used for regions that do not correspond to a gene region or Ensembl ID
    3 - Center  One or more genome sequencing center reporting the variant
    4 - NCBI_Build  The reference genome used for the alignment (GRCh38)
    5 - Chromosome  The affected chromosome (chr1)
    6 - Start_Position  Lowest numeric position of the reported variant on the genomic reference sequence. Mutation start coordinate
    7 - End_Position    Highest numeric genomic position of the reported variant on the genomic reference sequence. Mutation end coordinate
    8 - Strand  Genomic strand of the reported allele. Currently, all variants will report the positive strand: '+'
    9 - Variant_Classification  Translational effect of variant allele
    10 - Variant_Type   Type of mutation. TNP (tri-nucleotide polymorphism) is analogous to DNP (di-nucleotide polymorphism) but for three consecutive nucleotides. ONP (oligo-nucleotide polymorphism) is analogous to TNP but for consecutive runs of four or more (SNP, DNP, TNP, ONP, INS, DEL, or Consolidated)
    11 - Reference_Allele   The plus strand reference allele at this position. Includes the deleted sequence for a deletion or "-" for an insertion
    12 - Tumor_Seq_Allele1  Primary data genotype for tumor sequencing (discovery) allele 1. A "-" symbol for a deletion represents a variant. A "-" symbol for an insertion represents wild-type allele. Novel inserted sequence for insertion does not include flanking reference bases
    13 - Tumor_Seq_Allele2  Tumor sequencing (discovery) allele 2
    14 - dbSNP_RS   The rs-IDs from the   dbSNP database, "novel" if not found in any database used, or null if there is no dbSNP record, but it is found in other databases
    15 - dbSNP_Val_Status   The dbSNP validation status is reported as a semicolon-separated list of statuses. The union of all rs-IDs is taken when there are multiple
    16 - Tumor_Sample_Barcode   Aliquot barcode for the tumor sample
    17 - Matched_Norm_Sample_Barcode    Aliquot barcode for the matched normal sample
    """

    COSMIC_CLP = prefix + "/38/CosmicCLP_MutantExport.tsv.gz"
    COSMIC_CLP_NCV = prefix + "/38/CosmicCLP_NCVExport.tsv.gz"
    COSMIC_CLINICAL = prefix + "/cellline_nci60_clinical_data.tsv"

    ids = []
    with open(COSMIC_CLINICAL) as f:
        for line in islice(f, 1, None):
            fields = line.split("\t")
            ids.append(fields[0])

    with open(maf_file, 'w') as o:
        with gzip.open(COSMIC_CLP, mode='rt', encoding='UTF-8') as f:
            _, _, samples_raw = read_CLP_extended(f, 38, ids)
        with gzip.open(COSMIC_CLP_NCV, mode='rt', encoding='UTF-8') as f:
            _, _, samples_raw_NCV = read_CLP_NCV(f, 38, ids)

        for sample in samples_raw.keys():
            for mutation in chain(samples_raw[sample], samples_raw_NCV[sample]):
                chrom, pos, x, y = mutation
                o.write("\t\t\tGRCh38\t{}\t{}\t{}\t+\t\tSNP\t{}\t{}\t{}\t\t\t{}\t{}\n".format(
                    chrom, pos, pos, x, y, y, sample, sample))


def analyze_nci60_samples(prefix):
    if False:
        MAF_file = prefix + '/data_mutations_extended.txt'
        with open(MAF_file) as f:
            MAF = f.read()
            # print(MAF)
            samples, processing_stats = read_MAF_extended(MAF, 37)

    COSMIC_CLP = prefix + "/38/CosmicCLP_MutantExport.tsv.gz"
    COSMIC_CLP_NCV = prefix + "/38/CosmicCLP_NCVExport.tsv.gz"
    COSMIC_CLINICAL = prefix + "/cellline_nci60_clinical_data.tsv"
    THRESHOLD = 0.0  # 0.05

    ids = []
    with open(COSMIC_CLINICAL) as f:
        for line in islice(f, 1, None):
            fields = line.split("\t")
            ids.append(fields[0])

    with gzip.open(COSMIC_CLP, mode='rt', encoding='UTF-8') as f:
        samples, processing_stats = read_CLP_extended(f, 38, ids)
    # print(processing_stats)

    with gzip.open(COSMIC_CLP_NCV, mode='rt', encoding='UTF-8') as f:
        samples_NCV, processing_stats_NCV = read_CLP_NCV(f, 38, ids)
    # print(processing_stats_NCV)

    for k in samples_NCV.keys():
        for i in samples_NCV[k].keys():
            samples[k][i] += samples_NCV[k][i]

    # for k in processing_stats_NCV.keys():
    #     processing_stats[k]['loaded'] += processing_stats_NCV[k]['loaded']
    #     processing_stats[k]['skipped'] += processing_stats_NCV[k]['skipped']

    def format_numbers(d):
        d['score'] = "{:6.4g}".format((d['score']))
        return d

    # print(samples)
    # return

    for sample, mutations in samples.items():
        SAMPLE_ID = str(sample)
        # if sample != "786_0":
        #     break

        mutational_profile = get_mutational_profile(mutations)
        mutational_profile_counts = get_mutational_profile(mutations, counts=True)
        query_signature = make_bar_struct_from_values(mutational_profile)
        query_formatted = format_profile(mutational_profile)

        oname = prefix + "/profiles/{}.profile".format(SAMPLE_ID)
        print("ONAME:", oname)
        with open(oname, 'w') as o:
            query_formatted = format_profile(mutational_profile)
            o.write(query_formatted)

        oname = prefix + "/profiles/{}.counts".format(SAMPLE_ID)
        with open(oname, 'w') as o:
            query_formatted = format_profile(mutational_profile_counts, counts=True)
            o.write(query_formatted)

        oname = prefix + "/profiles/{}.stats".format(SAMPLE_ID)
        # print("ONAME:", oname)
        with open(oname, 'w') as o:
            o.write("N\t{}\n".format(int(sum(mutational_profile_counts))))
            o.write("Processed_C\t{}\n".format(processing_stats[sample]['processed']))
            o.write("Processed_NC\t{}\n".format(processing_stats_NCV[sample]['processed']))
            o.write("N_C\t{}\n".format(processing_stats[sample]['loaded']))
            o.write("N_NC\t{}\n".format(processing_stats_NCV[sample]['loaded']))
            o.write("Skipped_C\t{}\n".format(processing_stats[sample]['skipped']))
            o.write("Skipped_NC\t{}\n".format(processing_stats_NCV[sample]['skipped']))
            o.write("Skipped_indels_C\t{}\n".format(processing_stats[sample]['skipped_indels']))
            o.write("Skipped_indels_NC\t{}\n".format(processing_stats_NCV[sample]['skipped_indels']))
            o.write("Skipped_nucleotide_C\t{}\n".format(processing_stats[sample]['skipped_nucleotide']))
            o.write("Skipped_nucleotide_NC\t{}\n".format(processing_stats_NCV[sample]['skipped_nucleotide']))
            o.write("Skipped_context_C\t{}\n".format(processing_stats[sample]['skipped_context']))
            o.write("Skipped_context_NC\t{}\n".format(processing_stats_NCV[sample]['skipped_context']))

        classification_method = "rf"
        cancer_type_matches = [format_numbers(d) for d in identify_cancer_type(mutational_profile, classification_method)]
        primary_site_matches = [format_numbers(d) for d in identify_primary_site(mutational_profile, classification_method)]

        for method in "LJRZ":
        # for method in "D":
            # if method == 'M':
            #     _, contributing_signatures_A = decompose_mutational_profile(mutational_profile, "CLA")
            #     _, contributing_signatures_B = decompose_mutational_profile(mutational_profile, "CLB")
            #     _, contributing_signatures_cosmic = decompose_mutational_profile(mutational_profile, "IS")
            if method == 'L':
                _, _, contributing_signatures_A = decompose_mutational_profile_counts(mutational_profile_counts, "CLA", 'MLE', others_threshold=THRESHOLD)
                _, _, contributing_signatures_B = decompose_mutational_profile_counts(mutational_profile_counts, "CLB", 'MLE', others_threshold=THRESHOLD)
                _, _, contributing_signatures_cosmic = decompose_mutational_profile_counts(mutational_profile_counts, "IS", 'MLE', others_threshold=THRESHOLD)
            if method == 'J':
                _, _, contributing_signatures_A = decompose_mutational_profile_counts(mutational_profile_counts, "CLA", 'js')
                _, _, contributing_signatures_B = decompose_mutational_profile_counts(mutational_profile_counts, "CLB", 'js')
                _, _, contributing_signatures_cosmic = decompose_mutational_profile_counts(mutational_profile_counts, "IS", 'js')
            if method == 'R':
                _, _, contributing_signatures_A = decompose_mutational_profile_counts(mutational_profile_counts, "CLA", 'frobenius')
                _, _, contributing_signatures_B = decompose_mutational_profile_counts(mutational_profile_counts, "CLB", 'frobenius')
                _, _, contributing_signatures_cosmic = decompose_mutational_profile_counts(mutational_profile_counts, "IS", 'frobenius')
            if method == 'Z':
                _, _, contributing_signatures_A = decompose_mutational_profile_counts(mutational_profile_counts, "CLA", 'frobeniuszero')
                _, _, contributing_signatures_B = decompose_mutational_profile_counts(mutational_profile_counts, "CLB", 'frobeniuszero')
                _, _, contributing_signatures_cosmic = decompose_mutational_profile_counts(mutational_profile_counts, "IS", 'frobeniuszero')

            if method == 'D':
                _, _, contributing_signatures_cosmic = deconstruct_sigs(prefix + "/profiles/{}.profile".format(SAMPLE_ID), SAMPLE_ID)

            if method != 'D':
                oname = prefix + "/decompose/{}_{}-5.txt".format(SAMPLE_ID, method)
                with open(oname, 'w') as o:
                    for v in contributing_signatures_A:
                        o.write("{}\t{}\n".format(v['name'], v['score']))
                    # o.write("residuals\t{}\n".format(residuals_A))

                oname = prefix + "/decompose/{}_{}-10.txt".format(SAMPLE_ID, method)
                with open(oname, 'w') as o:
                    for v in contributing_signatures_B:
                        o.write("{}\t{}\n".format(v['name'], v['score']))
                    # o.write("residuals\t{}\n".format(residuals_B))

            oname = prefix + "/decompose/{}_{}-30.txt".format(SAMPLE_ID, method)
            with open(oname, 'w') as o:
                for v in contributing_signatures_cosmic:
                    o.write("{}\t{}\n".format(v['name'], v['score']))
                # o.write("residuals\t{}\n".format(residuals_cosmic))

    print(processing_stats)


if __name__ == '__main__':
    export_maf(
        prefix="/Users/gonceare/projects/Mutations/data/NCI60",
        maf_file="/Users/gonceare/projects/Mutations/data/NCI60/nci60.maf")

    # analyze_nci60_samples(prefix="/Users/gonceare/projects/Mutations/data/NCI60")

    # make_NCI60_report(prefix="/Users/gonceare/projects/Mutations/data/NCI60", n=30, decomposition='L')
    # make_NCI60_report(prefix="/Users/gonceare/projects/Mutations/data/NCI60", n=10, decomposition='L')
    # make_NCI60_report(prefix="/Users/gonceare/projects/Mutations/data/NCI60", n=5, decomposition='L')



import sys
import logging
from mutagene.io.fetch import fetch_genome, fetch_cohorts, fetch_examples, fetch_MSKCC


logger = logging.getLogger(__name__)
genome_error_message = 'requires genome name argument -g hg19, hg38, mm10, see http://hgdownload.cse.ucsc.edu/downloads.html for more'


class FetchMenu(object):
    def __init__(self, parser):
        parser.description = 'Download data from remote repositories and API'
        subparsers = parser.add_subparsers(
            dest='resource',
            title='subcommands',
            description='Choose data source',
            help='additional help available for subcommands')
        subparsers.required = True

        subparsers.add_parser('examples', description="""
This command will download sample1.maf and sample2.vcf files from the MutaGene server.
These files are only needed to reproduce examples in the manual.""")

        cohorts_parser = subparsers.add_parser('cohorts', help='cohorts help', description="""
This command will download mutational profiles and counts of observed mutations for cohorts available in online repositories.
The files are downloaded in one bundle cohorts.tar.gz that will be saved in the current directory.
Currently, only cohorts representing cancer types in COSMIC are provided.
Cohorts are required for ranking of mutations, because ranking relies upon counts of observed mutations and cancer type-specific profiles.
""")
        cohorts_parser.add_argument('source', nargs='?', choices=('COSMIC', 'GDC', 'MSKCC', 'ICGC'), default='COSMIC', help='Choose source data repository source (COSMIC by default)')
        cohorts_parser_action = cohorts_parser.add_mutually_exclusive_group(required=False)
        cohorts_parser_action.add_argument('--list', '-l', action='store_true', help='List available MSKCC cohorts')
        cohorts_parser_action.add_argument('--cohort', '-c', type=str, help='Specify MSKCC cohort')

        genome_parser = subparsers.add_parser('genome', description="""
This command will download reference genome assembly sequence in 2bit format from the UCSC genome browser website.

You need to specify the name of genome assembly in --genome (-g) argument.

Partial download is supported: if the process is interrupted run the same command again to continue downloading.
""")
        genome_parser.add_argument('--genome', '-g', type=str, help='hg38, hg19, mm10 according to UCSC genome browser nomenclature', required=True)

    @classmethod
    def examples(cls, args):
        # print('Cohorts', args)
        if args.resource == 'examples':
            fetch_examples()
            logger.info("Example files saved to current directory")

    @classmethod
    def cohorts(cls, args):
        # print('Cohorts', args)
        getattr(cls, 'cohorts_' + args.source)(args)

    @classmethod
    def cohorts_COSMIC(cls, args):
        fetch_cohorts()
        logger.info("cohorts.tar.gz saved to current directory")

    @classmethod
    def cohorts_GDC(cls, args):
        # fetch_cohorts()
        logger.warning("GDC currently not supported")

    @classmethod
    def cohorts_MSKCC(cls, args):
        if args.list:
            import pandas as pd
            with pd.option_context('display.max_rows', None):  # , 'display.max_columns', None
                print(pd.read_csv("https://www.cbioportal.org/webservice.do?cmd=getCancerStudies", sep="\t"))
        elif args.cohort:
            logger.info("Will download cohort " + args.cohort)
            fetch_MSKCC(args.cohort)
        else:
            logger.warning("Cohort not specified, use --list to display all available data sets")

    @classmethod
    def cohorts_ICGC(cls, args):
        # fetch_cohorts()
        logger.warning("ICGC currently not supported")

    @classmethod
    def genome(cls, args):
        if args.resource == 'genome':
            if not args.genome:
                logger.warning(genome_error_message)
                return
            fetch_genome(args.genome)
            logger.info("Twobit file saved to current directory")

    def callback(self, args):
        if not args.resource:
            logger.warning('No resource specified')
            sys.exit(1)

        getattr(self, args.resource)(args)



import glob
import random
import uuid
import numpy as np

from multiprocessing import Pool
from sklearn.metrics import (
    recall_score, precision_score, accuracy_score, f1_score, mean_squared_error)

from mutagene.io.profile import read_profile_file, write_profile, read_signatures
from mutagene.signatures.identify import NegLogLik
from mutagene.benchmark.deconstructsigs import deconstruct_sigs_custom
from mutagene.benchmark.generate_benchmark import *
# from mutagene.identify import decompose_mutational_profile_counts


def multiple_benchmark_helper(j):
    dirname = "data/benchmark/multiple"

    # for i in [5, 10, 30]:
    for i in [30, ]:
        W, signature_names = read_signatures(i)
        N = W.shape[1]

        # r = random.randrange(2, i // 3 + 2)
        r = random.randrange(2, min(i + 1, 15))

        # print(np.random.choice(N, r), .05 + np.random.dirichlet(np.ones(r), 1))
        while True:
            h0 = np.zeros(N)
            h0[np.random.choice(N, r)] = 0.05 + np.random.dirichlet(np.ones(r), 1)
            if np.greater(h0, 0.05).sum() == r:
                break
        h0 /= h0.sum()
        v0 = W.dot(h0)
        # print(h0)
        n_mutations = random.randrange(10, 50)
        v0_counts = np.random.multinomial(n_mutations, v0 / v0.sum())
        # print(v0_counts)

        random_name = str(uuid.uuid4())[:4]
        fname = dirname + "/{:02d}_{}_{}_{}".format(i, r, n_mutations, random_name)
        print(fname)
        profile_fname = fname + ".profile"
        info_fname = fname + ".info"
        mle_info = fname + ".MLE.info"
        mlez_info = fname + ".MLEZ.info"
        ds_info = fname + ".ds.info"

        write_profile(profile_fname, v0_counts)
        write_decomposition(info_fname, h0, signature_names)

        ##################################################
        results = deconstruct_sigs_custom(profile_fname, signatures=i)
        write_decomposition(ds_info, results, signature_names)
        ##################################################
        profile = read_profile_file(profile_fname)
        for method, method_fname in [("MLE", mle_info), ("MLEZ", mlez_info)]:
            _, _, results = decompose_mutational_profile_counts(
                profile,
                (W, signature_names),
                method,
                debug=False,
                others_threshold=0.0)
            write_decomposition(method_fname, results, signature_names)


def multiple_benchmark():
    # pathlib.Path(dirname).mkdir(parents=True, exist_ok=True)
    random.seed(13425)

    with Pool(10) as p:
        p.map(multiple_benchmark_helper, range(100))


def multiple_benchmark_run_helper(data):
    fname, signature_ids, W, force = data
    # methods = ['MLE', 'MLEZ', 'AICc', 'BIC', 'AICcZ', 'BICZ']
    methods = ['AICc', 'AICcZ']

    # print(fname)
    profile = read_profile_file(fname)

    for method in methods:
        info = "{}.{}.info".format(fname.split(".")[0], method)
        if isfile(info) and not force:
            continue

        print(info)

        _, _, results = decompose_mutational_profile_counts(
            profile,
            (W, signature_ids),
            method,
            debug=False,
            others_threshold=0.0)
        exposure_dict = {x['name']: x['score'] for x in results}
        exposure = [exposure_dict[name] for name in signature_ids]
        write_decomposition(info, np.array(exposure), signature_ids)


def multiple_benchmark_run(N, signature_ids, W, force=False):
    def get_iterator():
        for fname in glob.glob("data/benchmark/multiple/{:02d}_*.profile".format(N), recursive=True):
                yield (fname, signature_ids, W, force)

    random.seed(13425)
    with Pool(10) as p:
        p.map(multiple_benchmark_run_helper, get_iterator(), 100)


def aggregate_multiple_benchmarks():
    methods = {
        "mle": ".MLE.info",
        "mlez": ".MLEZ.info",
        "ds": ".ds.info",
        'aicc': '.AICc.info',
        'bic': '.BIC.info',
        'aiccz': '.AICcz.info',
        'bicz': '.BICz.info',
    }

    # signatures_thresholds = {
    #     5: 0.06,
    #     10: 0.03,
    #     30: 0.01,
    # }

    signatures_thresholds = {
        5: 0.06,
        10: 0.06,
        30: 0.06,
    }

    # signatures_thresholds = {
    #     5: 0.0001,
    #     10: 0.0001,
    #     30: 0.0001,
    # }

    # only report the signature 2 value (as in DeconstructSigs benchmark)
    with open("data/benchmark/multiple/res1.txt", 'w') as o:
        o.write("file_id\tsigtype\tnsig\tnmut\tmethod\tSRMSE\tPRMSE\tSTRMSE\tLLIK\tLLIK0\tTLLIK\tTLLIK0\tprecision\trecall\taccuracy\tf1\n")
        for fname in glob.glob("data/benchmark/multiple/*.profile", recursive=True):
            file_id = fname.split("/")[-1].split(".")[0]
            sigtype, r, nmut, replica = fname.split("/")[-1].split(".")[0].split("_")
            sigtype = int(sigtype)

            if sigtype != 30:
                continue

            W, signature_names = read_signatures(sigtype)

            info_fname = fname.split(".")[0] + '.info'
            orig_profile = read_profile_file(fname)
            h0, names = read_decomposition(info_fname)

            # threshold = 0.06
            threshold = 0.06

            # threshold = 1.0 / np.sqrt(int(nmut)) if method != "ds" else 0.06
            h0_threshold = np.where(h0 > threshold, h0, 0.0)  # zero below threshold
            h0_binary = np.array(h0_threshold) > 0.0   # true / false for threshold
            nsig = np.count_nonzero(h0_binary)

            if nsig < int(r):
                print("LESS", sigtype, nsig, r)

            if nsig > int(r):
                print("MORE", sigtype, nsig, r)

            if nsig <= 1:
                continue
            if nsig > 10:
                continue

            for method in methods:
                method_fname = fname.split(".")[0] + methods[method]
                values, names = read_decomposition(method_fname)

                # print(method_fname)

                if values is None:
                    continue

                h = np.array(values)
                if h.sum() == 0:
                    continue

                h_threshold = np.where(h > threshold, h, 0.0)  # zero below threshold

                reconstructed_profile = W.dot(h)
                # print(h)
                # print(reconstructed_profile)

                PRMSE = np.sqrt(mean_squared_error(
                    np.array(orig_profile) / np.array(orig_profile).sum(),
                    np.array(reconstructed_profile) / np.array(reconstructed_profile).sum()))
                SRMSE = np.sqrt(mean_squared_error(h0, h))
                STRMSE = np.sqrt(mean_squared_error(h0_threshold, h_threshold))
                LLIK0 = - NegLogLik(h0, W, orig_profile)
                TLLIK0 = - NegLogLik(h0_threshold, W, orig_profile)
                LLIK = - NegLogLik(h, W, orig_profile)
                TLLIK = - NegLogLik(h_threshold, W, orig_profile)

                # print(h0.sum())
                # print(h.sum())

                h_binary = np.array(h_threshold) > 0.0  # true / false for threshold
                precision = precision_score(h0_binary, h_binary)
                recall = recall_score(h0_binary, h_binary)
                accuracy = accuracy_score(h0_binary, h_binary)
                f1 = f1_score(h0_binary, h_binary)

                o.write("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(
                    file_id, sigtype, nsig, nmut, method, SRMSE, PRMSE, STRMSE, LLIK, LLIK0, TLLIK, TLLIK0, precision, recall, accuracy, f1))



import pandas as pd

from collections import defaultdict, OrderedDict
from functools import lru_cache, reduce

from scipy.stats import binom_test
from statsmodels.stats.multitest import multipletests

from mutagene.dna import *
from mutagene.io.profile import get_profile_attributes_dict

import logging
logger = logging.getLogger(__name__)


THRESHOLD_DRIVER = 8.030647e-05  # Max MCC
THRESHOLD_PASSENGER = 0.003440945  # 10% FPR


def get_all_codon_substitutions(P, seq5, Q):
    """
        Find all possible single nucleotide substitutions that mutate seq5 codon + context to amino acid Q
    """
    seq5_codon = seq5[1:-1]
    mutated_seq5s = []
    if seq5_codon not in codon_table:
        logger.warning("Invalid codon " + seq5_codon)
    if codon_table[seq5_codon] != P:
        logger.warning("Codon P does not match the sequence " + seq5_codon + " " + P)
    for codon, A in codon_table.items():
        if A != Q:
            continue
        n_changes = 0
        for i in range(3):
            if codon[i] != seq5_codon[i]:
                n_changes += 1
        if n_changes == 1:
            mutated_seq5s.append(seq5[0] + codon + seq5[-1])
    return mutated_seq5s


@lru_cache(maxsize=1000)
def predict_driver(observed, N, p, threshold_driver, threshold_passenger):
    """
    Binomial P-value based on the number of observed mutations, number of samples and expected mutability
    """
    # TODO: thresholds should be defined in the config
    # need more tumor samples than mutations for sure

    if N > 0 and observed > 0 and observed <= N:
        pvalue = binom_test(observed, N, p, alternative='greater')
    else:
        pvalue = 1.0

    if pvalue <= threshold_driver:
        prediction = "Driver"
        prediction_value = 1
    elif pvalue <= threshold_passenger:
        prediction = "Potential driver"
        prediction_value = 0.5
    elif pvalue < 1.0:
        prediction = "Passenger"
        prediction_value = 0
    else:
        prediction = "Undefined"
        prediction_value = -1

    return pvalue, prediction, prediction_value


def calculate_codon_mutability(mutation_model, seq5, mutated_seq5s):
    positional_mutability = defaultdict(float)
    for mutated_seq5 in mutated_seq5s:
        j = 0
        y = None
        for i in range(3):
            j = i + 1
            y = mutated_seq5[j]
            if seq5[j] != y:
                break
        if j > 0 and y is not None:
            # should always be > 0
            tri = seq5[j - 1: j + 2]
            positional_mutability[i] += mutation_model[tri][y]
    return 1.0 - reduce(lambda x, y: x * (1.0 - y), positional_mutability.values(), 1.0)


def calculate_base_substitution_mutability(counts_profile, cohort_size):
    assert cohort_size > 0
    assert len(counts_profile) == 96

    mutability = defaultdict(dict)
    counts_profile_dict = defaultdict(dict)

    for i, attrib in enumerate(get_profile_attributes_dict()):
        x, y = attrib['mutation']
        p5, p3 = attrib['context']
        tri = p5 + x + p3
        counts_profile_dict[tri][y] = counts_profile[i]

    for tri in counts_profile_dict.keys():
        tri_sum = float(sum(counts_profile_dict[tri].values()))
        mutability_tri = tri_sum / (float(cohort_size) * float(exome_trinucleotides[tri]))
        for y, y_counts in counts_profile_dict[tri].items():
            f = 0.0
            if tri_sum > 0:
                f = float(y_counts) / tri_sum
            mutability[tri][y] = f * mutability_tri
            mutability[complementary_trinucleotide[tri]][complementary_nucleotide[y]] = mutability[tri][y]

    return mutability


# import pprint
def rank(mutations_to_rank, outfile, profile, cohort_aa_mutations, cohort_size, threshold_driver, threshold_passenger):
    # profile_dict = profile_to_dict(profile)
    mutation_model = calculate_base_substitution_mutability(profile, cohort_size)
    cohort_size_corrected = cohort_size + 1  # add currently analyzed sample (many samples?) to cohort
    # pprint.pprint(mutation_model)
    results = []
    for mutation_key, mutation_value in mutations_to_rank.items():
        gene, mut = mutation_key
        seq5 = mutation_value['seq5']
        P, Q = mut[0], mut[-1]
        all_substitutions = get_all_codon_substitutions(P, seq5, Q)
        mutability = calculate_codon_mutability(mutation_model, seq5, all_substitutions)

        # print(mutation, mutability)

        # Observed k with pseudocount:
        observed_k = cohort_aa_mutations[gene].get(mut, 0)  # observed in precalculated cohort
        observed_k_pseudocount = observed_k + 1  # we also observe this mutation in analyzed sample

        pval, label, *_ = predict_driver(observed_k_pseudocount, cohort_size_corrected, mutability, threshold_driver, threshold_passenger)
        # print(gene, mut, seq5, all_substitutions, " = ", mutability, observed_k, cohort_size, " = ", pval, label)
        results.append({
            'gene': gene,
            'mutation': mut,
            'mutability': mutability,
            'observed': observed_k_pseudocount,
            'bscore': pval,
            'qvalue': pval,  # temporarily set to pval
            'label': label,
        })
    pvalues = [result['bscore'] for result in results]
    qvalues = []
    if len(pvalues):
        # qvalues = multipletests(pvals=pvalues, method='')[1]
        qvalues = multipletests(pvals=pvalues, method='fdr_bh')[1]
    for i, qvalue in enumerate(qvalues):
        results[i]['qvalue'] = qvalue

    results = list(map(OrderedDict, sorted(results, key=lambda k: k['bscore'])))
    if len(results) == 0:
        return
    df = pd.DataFrame(results, columns=results[0].keys())
    df.drop(df[df.mutability == 0].index, inplace=True)
    try:
        df.to_csv(outfile, sep="\t", index=False, doublequote=False)
    except (BrokenPipeError, IOError):
        pass



import csv
from collections import namedtuple

import twobitreader as tbr

from tqdm import tqdm
from collections import defaultdict
from itertools import cycle

from mutagene.dna import chromosome_name_mapping
from mutagene.motifs import nucleotides, complementary_nucleotide

import logging
logger = logging.getLogger(__name__)


def get_context_twobit_window(mutations, twobit_file, window_size):
    """
    User twobitreader to get context of mutations
    It's a special data structure.
    contexts[(chrom, pos)] = (nuc5, nuc3), seq_with_coords
    where seq_with_coords = [(chrom, pos, nucleotide, strand)]
    """
    if window_size is None:
        window_size = 50

    contexts = {}

    fname = twobit_file if twobit_file.endswith('.2bit') else twobit_file + '.2bit'
    f = tbr.TwoBitFile(fname)

    cn = complementary_nucleotide
    for (chrom, pos, transcript_strand, x, y) in mutations:
        start = int(pos) - 1  # 2bit uses zero-based numbering
        chrom = str(chrom)

        chromosome = chrom if chrom.startswith('chr') else 'chr' + chrom
        chromosome = chromosome_name_mapping.get(chromosome, chromosome)

        if chromosome in f:
            try:
                seq = f[chromosome][start - window_size:
                                    start + window_size + 1]
                assert len(seq) == window_size * 2 + 1
                seq = seq.upper()
            except Exception as e:
                logger.warning("TwoBit exception while reading the genome in {}:{}: {}".format(chrom, pos, e))
                continue
        else:
            logger.warning("Chromosome {} not found in 2bit file. Consider renaming it or using a different genome assembly".format(chromosome))
            continue

        strand = transcript_strand
        seq_with_coords = list(zip(
            cycle([chrom]),
            range(pos - window_size, pos + window_size + 1),
            seq,
            cycle([strand])))

        assert len(seq_with_coords) == len(seq)

        nuc5 = seq_with_coords[window_size - 1][2]
        nuc = seq_with_coords[window_size][2]
        nuc3 = seq_with_coords[window_size + 1][2]

        if nuc != 'N' and nuc != x:
            if cn[nuc] == x:
                nuc3 = cn[nuc5]
                nuc5 = cn[nuc3]
                # print('debug: complementary REF sequence detected')
            else:
                # print("{}:{}  {}>{}   {}[{}]{}".format(chromosome, pos, x, y, nuc5, nuc, nuc3))
                nuc3 = nuc5 = 'N'
            logger.warning(
                "REF allele does not match the genomic sequence in {}:{} {}!={}. Multiple errors could mean wrong genome assembly choice".format(
                    chromosome, pos, x, nuc))
        contexts[(chrom, pos)] = (nuc5, nuc3), seq_with_coords
    return contexts


def read_TCGI_with_context_window(infile, asm, window_size):
    """
    Tabular file; no particular column order required but must contain header line with four mandatory column names:
    (CHR, POS, REF, ALT) corresponding to the chromosome, position, reference and alternate allele columns, respectively
    • Optionally, if column name SAMPLE exists, the column is used as sample names, otherwise it is assumed that all variants are in the same sample
    • Optionally, if column name STRAND exists, the column is used as strandedness of variants
    (possible values: 1 or + for forward strand, -1 or - for reverse strand), otherwise it is assumed that all variants are described on the forward strand
    • The file can contain comment lines starting with #
    • In addition to the CHR, POS, REF and ALT columns, can have extra columns that will also appear in the final TCGI
    output table
    • Coordinates must follow the GRCh37 genome build, chromosome names are accepted either with or without the ‘chr’ prefix
    • Alleles can follow both the Ensembl or VCF convention (e.g. for a deletion, both ‘ATCA to A’ or ‘TCA to –’ forms are accepted)

    https://www.cancergenomeinterpreter.org/images/inputformats.pdf

    returns mutations, mutations_with_context, processing_stats
    """
    cn = complementary_nucleotide
    mutations = defaultdict(lambda: defaultdict(float))
    N_skipped = 0

    processing_stats = {'loaded': 0, 'skipped': 0, 'nsamples': 0, 'format': 'unknown'}
    if not infile:
        logger.warning("No input file")
        return mutations, {}, processing_stats

    try:
        reader = csv.reader((row for row in infile if not row.startswith('#')), delimiter='\t')
        # get names from column headers
        header = next(reader)
        header = tuple(map(lambda s: s.lower().replace('.', '_'), header))
        # print(header)
        TCGI = namedtuple("TCGI", header, rename=True)
    except ValueError:
        logger.warning("TCGI format not recognized")
        raise
        return mutations, {}, processing_stats

    raw_mutations = defaultdict(list)
    # for line in tqdm(infile):
    for data in tqdm(map(TCGI._make, reader), leave=False):
        # chromosome is expected to be one or two number or one letter
        if hasattr(data, 'chrom'):
            chrom = data.chrom
        else:
            raise ValueError('Chromosome is not defined in TCGI file')

        if hasattr(data, 'sample'):
            sample = data.sample
        else:
            sample = "TCGI"

        if hasattr(data, 'ref'):
            x = data.ref
        else:
            raise ValueError('Reference allele (REF) is not defined in TCGI file')

        if hasattr(data, 'alt'):
            y = data.alt
        else:
            raise ValueError('Variant allele is not defined in TCGI file')

        if y is None:
            continue
        # skip if found unexpected nucleotide characters
        if len(set([x, y]) - set(nucleotides)) > 0:
            continue

        if hasattr(data, 'pos'):
            try:
                pos = int(data.pos)
            except ValueError:
                raise ValueError('Start position is not a number in TCGI file')
        else:
            raise ValueError('Start position is not defined in TCGI file')

        transcript_strand = '+'
        raw_mutations[sample].append((chrom, pos, transcript_strand, x, y))

    mutations_with_context = defaultdict(list)

    for sample, sample_mutations in raw_mutations.items():
        if len(sample_mutations) > 0:
            contexts = get_context_twobit_window(sample_mutations, asm, window_size)

            if contexts is None or len(contexts) == 0:
                return None, None

            for (chrom, pos, transcript_strand, x, y) in sample_mutations:
                (p5, p3), seq_with_coords = contexts.get((chrom, pos), (("N", "N"), []))

                if len(set([p5, x, y, p3]) - set(nucleotides)) > 0:
                    # print("Skipping invalid nucleotides")
                    N_skipped += 1
                    continue

                if x in "CT":
                    mutations[sample][p5 + p3 + x + y] += 1.0
                else:
                    # complementary mutation
                    mutations[sample][cn[p3] + cn[p5] + cn[x] + cn[y]] += 1.0
                mutations_with_context[sample].append((chrom, pos, transcript_strand, x, y, seq_with_coords))

    N_loaded = 0
    for sample, sample_mutations in mutations.items():
        N_loaded += int(sum(sample_mutations.values()))

    processing_stats = {
        'loaded': N_loaded,
        'skipped': N_skipped,
        'nsamples': len(mutations.keys()),
        'format': 'TCGI'
    }
    return mutations, mutations_with_context, processing_stats


def read_mutations(file_format, *args, **kwargs):
    """ Wrapper for read_X_with_context_window """
    function_name = "read_{}_with_context_window".format(file_format)
    return globals()[function_name](*args, **kwargs)


def read_MAF_with_context_window(infile, asm, window_size):
    """
        Read MAF file and extract context of mutations for assembly asm and window +/- window_size around each mutation
        MAF format description: https://docs.gdc.cancer.gov/Data/File_Formats/MAF_Format/

        returns mutations, mutations_with_context, processing_stats
    """
    cn = complementary_nucleotide
    mutations = defaultdict(lambda: defaultdict(float))
    N_skipped = 0

    processing_stats = {'loaded': 0, 'skipped': 0, 'nsamples': 0, 'format': 'unknown'}
    if not infile:
        logger.warning("No input file")
        return mutations, {}, processing_stats

    try:
        reader = csv.reader((row for row in infile if not row.startswith('#')), delimiter='\t')
        # get names from column headers
        header = next(reader)
        header = tuple(map(lambda s: s.lower().replace('.', '_'), header))
        # print(header)
        MAF = namedtuple("MAF", header, rename=True)
    except ValueError:
        logger.warning("MAF format not recognized")
        raise
        return mutations, {}, processing_stats

    raw_mutations = defaultdict(list)
    # for line in tqdm(infile):
    for data in tqdm(map(MAF._make, reader), leave=False):
        # assembly_build = col_list[3]  # MAF ASSEMBLY

        # chromosome is expected to be one or two number or one letter
        if hasattr(data, 'chromosome'):
            chrom = data.chromosome  # MAF CHROM
        else:
            raise ValueError('Chromosome is not defined in MAF file')

        if hasattr(data, 'tumor_sample_barcode'):
            sample = data.tumor_sample_barcode       # Tumor barcode
        elif hasattr(data, 'sample_id'):
            sample = data.sample_id
        else:
            raise ValueError("Sample ID is not defined in MAF file")

        if hasattr(data, 'reference_allele'):
            x = data.reference_allele            # MAF REF
        else:
            raise ValueError('Reference allele is not defined in MAF file')

        if hasattr(data, 'variant_allele'):
            y = data.variant_allele
        elif hasattr(data, 'tumor_seq_allele1') and hasattr(data, 'tumor_seq_allele2'):
            y1 = data.tumor_seq_allele1           # MAF ALT1
            y2 = data.tumor_seq_allele2           # MAF ALT2
            y = y1 if y1 != x else None
            y = y2 if y2 != x else y
        else:
            raise ValueError('Variant allele is not defined in MAF file')

        if y is None:
            continue
        # skip if found unexpected nucleotide characters
        if len(set([x, y]) - set(nucleotides)) > 0:
            continue

        if hasattr(data, 'start_position'):
            try:
                pos = int(data.start_position)
            except ValueError:
                raise ValueError('Start position is not a number in MAF file')
        else:
            raise ValueError('Start position is not defined in MAF file')

        # assuming that reference strand for reported mutations is always '+'!

        # transcript strand could be anything
        if hasattr(data, 'transcript_strand'):
            transcript_strand = data.transcript_strand
            # GDC uses 1 and -1
            if transcript_strand == '+':
                pass
            elif transcript_strand == '-':
                pass
            elif transcript_strand == '1':
                transcript_strand = '+'
            elif transcript_strand == '-1':
                transcript_strand = '-'
            elif transcript_strand == '':
                transcript_strand = '+'  # default value
            else:
                raise ValueError('Unexpected value of transcript_strand in MAF file')
        else:
            # this is an incorrect assumption about transcript strand
            transcript_strand = '+'

        raw_mutations[sample].append((chrom, pos, transcript_strand, x, y))

    mutations_with_context = defaultdict(list)

    for sample, sample_mutations in raw_mutations.items():
        if len(sample_mutations) > 0:
            contexts = get_context_twobit_window(sample_mutations, asm, window_size)

            if contexts is None or len(contexts) == 0:
                return None, None

            for (chrom, pos, transcript_strand, x, y) in sample_mutations:
                (p5, p3), seq_with_coords = contexts.get((chrom, pos), (("N", "N"), []))

                if len(set([p5, x, y, p3]) - set(nucleotides)) > 0:
                    # print("Skipping invalid nucleotides")
                    N_skipped += 1
                    continue

                if x in "CT":
                    mutations[sample][p5 + p3 + x + y] += 1.0
                else:
                    # complementary mutation
                    mutations[sample][cn[p3] + cn[p5] + cn[x] + cn[y]] += 1.0
                mutations_with_context[sample].append((chrom, pos, transcript_strand, x, y, seq_with_coords))

    N_loaded = 0
    for sample, sample_mutations in mutations.items():
        N_loaded += int(sum(sample_mutations.values()))

    processing_stats = {
        'loaded': N_loaded,
        'skipped': N_skipped,
        'nsamples': len(mutations.keys()),
        'format': 'MAF'
    }

    return mutations, mutations_with_context, processing_stats


def read_VCF_with_context_window(infile, asm, window_size):
    """
    Read VCF file and extract context of mutations for assembly asm and window +/- window_size around each mutation
    returns mutations, mutations_with_context, processing_stats
    """
    cn = complementary_nucleotide
    mutations = defaultdict(lambda: defaultdict(float))
    mutations_with_context = defaultdict(list)
    raw_mutations = defaultdict(list)

    N_skipped = 0
    # N_skipped_indels = 0

    sample = 'VCF'

    for line in infile:
        if line.startswith("#"):
            continue
        if len(line) < 10:
            continue

        col_list = line.split()
        if len(col_list) < 4:
            continue

        # ID = col_list[2]

        # chromosome is expected to be one or two number or one letter
        chrom = col_list[0]  # VCF CHROM
        if chrom.lower().startswith("chr"):
            chrom = chrom[3:]
        # if len(chrom) == 2 and chrom[1] not in "0123456789":
        #     chrom = chrom[0]

        pos = int(col_list[1])  # VCF POS
        x = col_list[3]         # VCF REF
        y = col_list[4]         # VCF ALT

        # if multiple REF or ALT alleles are given, ignore mutation entry (could mean seq error, could mean deletion)
        if len(x) != 1:
            N_skipped += 1
            continue
        if len(y) != 1:
            N_skipped += 1
            continue

        transcript_strand = '+'
        raw_mutations[sample].append((chrom, pos, transcript_strand, x, y))
    # print("RAW", raw_mutations)
    # print("INDELS", N_skipped)

    for sample, sample_mutations in raw_mutations.items():
        if len(sample_mutations) > 0:
            contexts = get_context_twobit_window(sample_mutations, asm, window_size)
            if contexts is None or len(contexts) == 0:
                return None, None

            for (chrom, pos, transcript_strand, x, y) in sample_mutations:
                (p5, p3), seq_with_coords = contexts.get((chrom, pos), (("N", "N"), []))
                # print("RESULT: {} {}".format(p5, p3))

                if len(set([p5, x, y, p3]) - set(nucleotides)) > 0:
                    # print(chrom, pos, p5, p3, x)
                    # print("Skipping invalid nucleotides")
                    N_skipped += 1
                    continue

                if x in "CT":
                    mutations[sample][p5 + p3 + x + y] += 1.0
                else:
                    # complementary mutation
                    mutations[sample][cn[p3] + cn[p5] + cn[x] + cn[y]] += 1.0

                mutations_with_context[sample].append((chrom, pos, transcript_strand, x, y, seq_with_coords))

    N_loaded = 0
    for sample, sample_mutations in mutations.items():
        N_loaded += int(sum(sample_mutations.values()))

    nsamples = len(mutations.keys())
    processing_stats = {
        'loaded': N_loaded,
        'skipped': N_skipped,
        'format': 'VCF',
        'nsamples': nsamples
    }
    return mutations, mutations_with_context, processing_stats



import argparse
import sys
import logging
from pathlib import Path

from mutagene.mutability.mutability import rank, THRESHOLD_DRIVER, THRESHOLD_PASSENGER
from mutagene.io.cohorts import read_cohort_mutations_from_tar
from mutagene.io.cohorts import read_cohort_size_from_profile_file, list_cohorts_in_tar
from mutagene.io.profile import read_profile_file
from mutagene.io.protein_mutations_MAF import read_MAF_with_genomic_context


logger = logging.getLogger(__name__)
genome_error_message = """requires genome name argument -g hg19, hg38, mm10, see http://hgdownload.cse.ucsc.edu/downloads.html for more
                          Use mutagene fetch to download genome assemblies"""


class RankMenu(object):
    def __init__(self, parser):
        parser.add_argument("--infile", "-i", nargs='*', help="Input file in MAF format", type=argparse.FileType('r'))
        parser.add_argument('--outfile', "-o", nargs='?', type=argparse.FileType('w'), default=sys.stdout)
        parser.add_argument('--genome', "-g", help="Location of genome assembly file", type=str)

        # parser.add_argument('--mode', "-m", help="n (nucleotide) or aa (amino acid) mode", type=str, default="aa")

        parser.add_argument('--cohorts-file', type=str, help="Location of tar.gz container or directory for cohorts", default="cohorts.tar.gz")
        parser.add_argument('--cohort', "-c", type=str, help="Name of cohort with observed mutations")

        parser.add_argument('--profile', "-p", help="Override profile to calculate mutability, may also describe cohort size", type=str)
        parser.add_argument('--nsamples', "-n", type=int, help="Override cohort size")
        parser.add_argument('--threshold-driver', "-td", help="BScore threshold between Driver and Pontential Driver mutations", type=float, default=THRESHOLD_DRIVER)
        parser.add_argument('--threshold-passenger', "-tp", help="BScore threshold between Pontential Driver and Passenger mutations", type=float, default=THRESHOLD_PASSENGER)

    def callback(self, args):
        if not args.genome:
            logger.warning(genome_error_message)
            return

        if not Path(args.cohorts_file).is_file():
            logger.warning("Cohorts file missing. Download with \"mutagene fetch_cohorts\"")
            return

        if args.cohort and Path(args.cohorts_file).is_file():
            profile, cohort_size, cohort_aa_mutations, cohort_na_mutations = read_cohort_mutations_from_tar(args.cohorts_file, args.cohort)
            logger.info('Profile and cohort size loaded from precalculated cohorts N=' + str(cohort_size))
        else:
            logger.warning('Cohort required')

            if args.cohorts_file:
                logger.warning("List of available cohorts:\n" + list_cohorts_in_tar(args.cohorts_file))
            return

        if args.profile:
            profile = read_profile_file(args.profile)
            if profile:
                logger.info('Profile overridden')
            else:
                return
            cohort_size_new = read_cohort_size_from_profile_file(args.profile)
            if cohort_size_new:
                cohort_size = cohort_size_new
                logger.info('Cohort size loaded from profile N=' + str(cohort_size))

        if args.nsamples:
            cohort_size = args.nsamples
            logger.info('Cohort size overridden N=' + str(cohort_size))

        if isinstance(args.infile, list):
            if len(args.infile) > 1:
                logger.info('Multiple input files provided')
            mutations_to_rank = {}
            processing_stats = {'loaded': 0, 'skipped': 0}
            for infile in args.infile:
                mut, stats = read_MAF_with_genomic_context(infile, args.genome)
                mutations_to_rank.update(mut)
                processing_stats['loaded'] += stats['loaded']
                processing_stats['skipped'] += stats['skipped']
        else:
            mutations_to_rank, processing_stats = read_MAF_with_genomic_context(args.infile, args.genome)

        if not len(mutations_to_rank):
            logger.warning('MutaGene rank failed: No mutations to rank. Check that the infile is in MAF format')
            return

        msg = "Loaded {} mutations".format(processing_stats['loaded'])
        if processing_stats['skipped'] > 0:
            msg += " skipped {} mutations".format(processing_stats['skipped'])
        logger.info(msg)

        # mutations_to_rank = list(mutations_to_rank)

        # not_unique = len(mutations_to_rank)
        # mutations_to_rank = list(set(mutations_to_rank))
        # unique = len(mutations_to_rank)
        # if not_unique != unique:
        #     logger.info("Removed {} duplicate mutations".format(not_unique - unique))

        logger.info("THRESHOLD_DRIVER: {}".format(args.threshold_driver))
        logger.info("THRESHOLD_PASSENGER: {}".format(args.threshold_passenger))

        rank(mutations_to_rank, args.outfile, profile, cohort_aa_mutations, cohort_size, args.threshold_driver, args.threshold_passenger)


"""

# parser.add_argument('integers', metavar='N', type=int, nargs='+',
#                     help='an integer for the accumulator')
# parser.add_argument('--sum', dest='accumulate', action='store_const',
#                     const=sum, default=max,
#                     help='sum the integers (default: find the max)')
parser.add_argument("--infile", "-i", nargs='*', help="Input file format", type=argparse.FileType('r'))
parser.add_argument('--outfile', "-o", nargs='?', type=argparse.FileType('w'), default=sys.stdout)
# parser.add_argument('--signatures', type=str)
# parser.add_argument('--motifs', type=str)

# parser.add_argument('--threshold', "-t", help="B-score thresholds for drivers", type=float)
# parser.add_argument('--mut-rate', "-r", help="Mutation rate overrides the rate inferred from profile", type=float)

"""



import json
from subprocess import Popen, PIPE


def deconstruct_sigs(profile_fname, sample='sample'):
    script1 = """
library(jsonlite)
library(deconstructSigs)
"""
    script2 = """
s <- t(read.table('{}', sep="\t", header=FALSE, row.names=1))
row.names(s) <- '{}'
s <- as.data.frame(s)
w <- whichSignatures(s / sum(s), signatures.ref=signatures.cosmic, signature.cutoff=0.00)
toJSON(w)
"""
    script2 = (script1 + script2).format(profile_fname, sample).encode("utf-8")

    proc = Popen(["Rscript", "-"], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    out, err = proc.communicate(script2, timeout=10)
    # exitcode = proc.returncode
    # print(exitcode, out, err)
    json_string = out.decode("utf-8")
    w = json.loads(json_string)
    result = []
    for k, v in w['weights'][0].items():
        if k.startswith('_row'):
            continue
        # if float(v) == 0.0:
        #     continue
        name = k.replace('Signature.', '')
        result.append({
            'name': name,
            'score': v})
    return result


def deconstruct_sigs_custom(profile_fname, sample='sample', signatures=30, cutoff=0.00, reverse=False):
    script1 = """
library(jsonlite)
library(deconstructSigs)
"""
    script1 += "W <- data.frame()"

    sig_map = {
        5: "A",
        10: "B",
        30: "C"
    }

    if reverse:
        signatures_list = list(reversed(range(signatures)))
    else:
        signatures_list = list(range(signatures))

    for i in signatures_list:
        script1 += """
W <- rbind(W, t(read.table('/Users/agoncear/projects/mutagene/data/signatures/{}_{}.profile', sep="\t", header=FALSE, row.names=1)))
""".format(sig_map[signatures], i + 1)

    script1 += "row.names(W) <- c(" + ",".join(["'Signature." + str(i + 1) + "'" for i in signatures_list]) + ")\n"
    script1 += "W <- as.data.frame(W)\n"

    script2 = """
s <- t(read.table('{}', sep="\t", header=FALSE, row.names=1))
row.names(s) <- '{}'
s <- as.data.frame(s)
w <- whichSignatures(s / sum(s), signatures.ref=W, signature.cutoff={})
toJSON(w)
"""
    script2 = (script1 + script2).format(profile_fname, sample, cutoff)
    script2 = script2.encode("utf-8")

    proc = Popen(["Rscript", "-"], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    out, err = proc.communicate(script2, timeout=10)
    # exitcode = proc.returncode
    # print(exitcode, out, err)
    json_string = out.decode("utf-8")
    w = json.loads(json_string)
    # import pprint
    # pprint.pprint(w)
    result = []
    for k, v in w['weights'][0].items():
        if k.startswith('_row'):
            continue
        # if float(v) == 0.0:
        #     continue
        name = k.replace('Signature.', '')
        result.append({
            'name': name,
            'score': v})
    return result



import requests
import json
import os
import sys

import logging
logger = logging.getLogger(__name__)


def gdc_read_file(file_id="11443f3c-9b8b-4e47-b5b7-529468fec098"):
    data_endpt = "https://api.gdc.cancer.gov/slicing/view/{}".format(file_id)

    TOKEN_FILE_PATH = os.environ.get('GDC_TOKEN')
    if not TOKEN_FILE_PATH:
        logger.warning("GDC_TOKEN environment variable should point to GDC token file")
        sys.exit(1)
    with open(TOKEN_FILE_PATH, "r") as token:
        token_string = str(token.read().strip())

    params = {"gencode": ["BRCA1", "BRCA2"]}

    response = requests.post(
        data_endpt,
        data=json.dumps(params),
        headers={
            "Content-Type": "application/json",
            "X-Auth-Token": token_string
        })
    return response.content



import tarfile
from collections import defaultdict
from mutagene.io.profile import read_profile_str

import logging
logger = logging.getLogger(__name__)


def read_cohort_size_from_profile_file(profile_file):
    try:
        with open(profile_file) as f:
            profile_str = f.read()
            return read_cohort_size_from_profile_str(profile_str)
    except IOError:
        logger.warning("Could not read profile file")
        return 0


def read_cohort_size_from_profile_str(profile_str):
    for line in profile_str.splitlines():
        if line.startswith("#"):
            # print(line)
            a, b = line.strip().split()
            if a == "#NSAMPLES":
                return int(b)
    return 0


def list_cohorts_in_tar(tar_fname):
    """ Returns a multiline string formatted list of cohorts contained in tar file """
    cohorts = []
    with tarfile.open(tar_fname, 'r:gz') as tar:
        for t in tar:
            haystack = t.name.lower()
            if haystack.endswith(".aa_mutations.txt"):
                cohorts.append("\t" + haystack.split("/")[1].split('.')[0])
    return "\n".join(cohorts)


def read_aa_mutations_map(aa_str):
    mutations = defaultdict(dict)
    for line in aa_str.splitlines():
        if len(line) == 0:
            continue
        fields = line.split()
        if len(fields) != 3:
            continue
        gene, mut, count = fields
        count = int(count)
        mutations[gene][mut] = count
    return mutations


def read_na_mutations_map(na_str):
    mutations = defaultdict(dict)
    for line in na_str.splitlines():
        if len(line) == 0:
            continue
        fields = line.split()
        if len(fields) != 5:
            continue
        chrom, pos, ref, alt, count = fields
        count = int(count)
        mutations[chrom][(pos, ref, alt)] = count
    return mutations


def read_cohort_mutations_from_tar(tar_fname, cohort):
    """ Loads up profile, cohort size, aa mutations and na mutations from precalculated cohorts tar file """
    aa_mutations = {}
    na_mutations = {}
    profile = []
    cohort_size = 0
    with tarfile.open(tar_fname, 'r:gz') as tar:
        for t in tar:
            haystack = t.name.lower()
            needle = "/{}.".format(cohort.lower())
            if haystack.find(needle) != -1:
                if haystack.endswith('.profile'):
                    profile_str = tar.extractfile(t).read().decode('utf-8')
                    profile = read_profile_str(profile_str)
                    cohort_size = read_cohort_size_from_profile_str(profile_str)
                if haystack.endswith('.aa_mutations.txt'):
                    aa_mutations_str = tar.extractfile(t).read().decode('utf-8')
                    aa_mutations = read_aa_mutations_map(aa_mutations_str)
                if haystack.endswith('.dna_mutations.txt'):
                    na_mutations_str = tar.extractfile(t).read().decode('utf-8')
                    na_mutations = read_na_mutations_map(na_mutations_str)
    return profile, cohort_size, aa_mutations, na_mutations
